/*
 * Copyright (c) 2017-2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Feb 27 2017
 */
#include "acheader.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "arduio.h"


/* --- local data ---------------------------------------------------------- */

struct ario_iobuffer {
  volatile uint8_t recvbuf[ARDUIO_RECV_BUF];
  volatile uint8_t recv_r, recv_w;

  struct {
    unsigned start:1;
    unsigned escape:1;
  } state;
  struct ario_msg msg;

  /* send buffer must be large enough for ddir + dpin + apin states, i.e.: */
#if 9 + 3 * ARIO_DPINS + 2 * ARIO_APINS > ARDUIO_SEND_BUF
#error "send buffer too small"
#endif
  volatile uint8_t sendbuf[ARDUIO_SEND_BUF];
  volatile uint8_t send_r, send_w;
};

static struct ario_iobuffer uart;

static void	ario_encodechar(char c);
static void	ario_putchar(char c);
static void	ario_getchar(char c);


/* --- ario_uart_init ------------------------------------------------------- */

void
ario_uart_init(uint32_t baud)
{
  uint16_t b = F_CPU / baud / 16;

  /* round to the nearest integer */
  if (F_CPU % (baud * 16) <= baud * 8)
    b--;

  /* initialize buffers */
  uart.recv_r = 0;
  uart.recv_w = 0;
  uart.send_r = 0;
  uart.send_w = 0;

  /* set baudrate */
  UBRR0 = b;

  /* normal clock speed */
  UCSR0A = (0 << U2X0);
  /* 8/N/1 */
  UCSR0C = (1 << UCSZ01) | (1 << UCSZ00);
  /* enable transmission and reception */
  UCSR0B = (1 << RXCIE0) | (1 << RXEN0) | (1 << TXEN0);
}


/* --- ario_log ------------------------------------------------------------- */

void
ario_log(const char *fmt, ...)
{
  va_list ap;
  char c;

  va_start(ap, fmt);
  ario_putchar('^');
  while((c = pgm_read_byte(fmt++))) {
    switch(c) {
      case '%': {
        switch(pgm_read_byte(fmt++)) {
          case '1':
            ario_encodechar(va_arg(ap, int));
            break;

          case '2': {
            uint16_t x = va_arg(ap, uint16_t);
            ario_encodechar((x >> 8) & 0xff);
            ario_encodechar(x & 0xff);
            break;
          }

          case '4': {
            uint32_t x = va_arg(ap, uint32_t);
            ario_encodechar((x >> 24) & 0xff);
            ario_encodechar((x >> 16) & 0xff);
            ario_encodechar((x >> 8) & 0xff);
            ario_encodechar(x & 0xff);
            break;
          }
        }
        break;
      }

      default:
        ario_encodechar(c);
    }
  }
  ario_putchar('$');
  va_end(ap);
}


/* --- ario_log_buffer ------------------------------------------------------ */

void
ario_log_buffer(const uint8_t *buffer, uint8_t len)
{
  ario_putchar('^');
  while(len--)
    ario_encodechar(*buffer++);
  ario_putchar('$');
}


/* --- ario_io_recv --------------------------------------------------------- */

struct ario_msg *
ario_recv(void)
{
  uint8_t r;
  char c;

  r = uart.recv_r;
  while(r != uart.recv_w) {
    c = uart.recvbuf[r];
    r = (r + 1) % sizeof(uart.recvbuf);
    uart.recv_r = r;

    switch (c) {
      case '^':
        uart.state.start = 1;
        uart.state.escape = 0;
        uart.msg.len = 0;
        break;

      case '$':
        if (!uart.state.start) break;
        uart.state.start = 0;
        return &uart.msg;

      case '!':
        uart.state.start = 0;
        break;

      case '\\':
        uart.state.escape = 1;
        break;

      default:
        if (!uart.state.start) break;
        if (uart.msg.len >= sizeof(uart.msg.data)) {
          uart.state.start = 0;
          break;
        }
        if (uart.state.escape) {
          uart.state.escape = 0;
          c = ~c;
        }

        uart.msg.data[uart.msg.len++] = c;
        break;
    }
  }

  return NULL;
}


/* --- ario_send ------------------------------------------------------------ */

void
ario_send(void)
{
  /* synchronous send */
  do {
    uint8_t w = uart.send_w;
    uint8_t r = uart.send_r;

    while (r != w) {
      while (!(UCSR0A & (1 << UDRE0)));
      UDR0 = uart.sendbuf[r];

      r = (r + 1) % sizeof(uart.sendbuf);
    }
    uart.send_r = r;
  } while(0);
}


/* --- ario_encodechar ------------------------------------------------------ */

static void
ario_encodechar(char c)
{
  switch (c) {
    case '^': case '$':  case '!': case '\\':
      ario_putchar('\\');
      c = ~c;
      break;
  }
  ario_putchar(c);
}


/* --- ario_putchar --------------------------------------------------------- */

static void
ario_putchar(char c)
{
  uint8_t w, wp;

  w = uart.send_w;
  wp = (w + 1) % sizeof(uart.sendbuf);

  if (wp == uart.send_r) {
    /* discard byte in case of buffer overflow */
    uart.sendbuf[w] = '!';
    return;
  }

  uart.sendbuf[w] = c;
  uart.send_w = wp;
}


/* --- ario_getchar --------------------------------------------------------- */

static inline  __attribute__((always_inline)) void
ario_getchar(char c)
{
  uint8_t w, wp;

  /* push byte */
  w = uart.recv_w;
  wp = (w + 1) % sizeof(uart.recvbuf);
  if (wp == uart.recv_r) {
    /* discard byte in case of buffer overflow */
    uart.recvbuf[w] = '!';
    return;
  }

  uart.recvbuf[w] = c;
  uart.recv_w = wp;
}


/* --- USART_RX_vect ------------------------------------------------------- */

inline __attribute__((always_inline)) void
ario_poll_uart(void)
{
  uint8_t status = UCSR0A;

  if (status & (1 << RXC0)) {
    uint8_t c = UDR0;

    if (status & ((1 << FE0) | (1 << DOR0)))
      c = '!';

    ario_getchar(c);
  }
}

#ifdef USART_RX_vect
ISR(USART_RX_vect)
#else
ISR(USART0_RX_vect)
#endif
{
  ario_poll_uart();
}
