/*
 * Copyright (c) 2017-2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Feb 27 2017
 */
#include "acheader.h"

#include <avr/interrupt.h>
#include <avr/io.h>

#include "arduio.h"
#include <util/atomic.h>
#include <util/delay.h>


static enum ario_dir ario_dpins[ARIO_DPINS];
static uint16_t ario_apins[ARIO_APINS];

static uint16_t ario_wspins[ARIO_DPINS][16];

/* --- pin configurations -------------------------------------------------- */

static const struct {
  volatile uint8_t *port;
  volatile uint8_t *ddr;
  volatile uint8_t *pin;
  uint8_t bit;
} dpinsreg[ARIO_DPINS] = {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
  [ 0] = { &PORTD, &DDRD, &PIND, 1<<0 },
  [ 1] = { &PORTD, &DDRD, &PIND, 1<<1 },
  [ 2] = { &PORTD, &DDRD, &PIND, 1<<2 },
  [ 3] = { &PORTD, &DDRD, &PIND, 1<<3 },
  [ 4] = { &PORTD, &DDRD, &PIND, 1<<4 },
  [ 5] = { &PORTD, &DDRD, &PIND, 1<<5 },
  [ 6] = { &PORTD, &DDRD, &PIND, 1<<6 },
  [ 7] = { &PORTD, &DDRD, &PIND, 1<<7 },
  [ 8] = { &PORTB, &DDRB, &PINB, 1<<0 },
  [ 9] = { &PORTB, &DDRB, &PINB, 1<<1 },
  [10] = { &PORTB, &DDRB, &PINB, 1<<2 },
  [11] = { &PORTB, &DDRB, &PINB, 1<<3 },
  [12] = { &PORTB, &DDRB, &PINB, 1<<4 },
  [13] = { &PORTB, &DDRB, &PINB, 1<<5 }

#elif defined(ARDUIO_MEGA2560)
  [ 0] = { &PORTE, &DDRE, &PINE, 1<<0 },
  [ 1] = { &PORTE, &DDRE, &PINE, 1<<1 },
  [ 2] = { &PORTE, &DDRE, &PINE, 1<<4 },
  [ 3] = { &PORTE, &DDRE, &PINE, 1<<5 },
  [ 4] = { &PORTG, &DDRG, &PING, 1<<5 },
  [ 5] = { &PORTE, &DDRE, &PINE, 1<<3 },
  [ 6] = { &PORTH, &DDRH, &PINH, 1<<3 },
  [ 7] = { &PORTH, &DDRH, &PINH, 1<<4 },
  [ 8] = { &PORTH, &DDRH, &PINH, 1<<5 },
  [ 9] = { &PORTH, &DDRH, &PINH, 1<<6 },
  [10] = { &PORTB, &DDRB, &PINB, 1<<4 },
  [11] = { &PORTB, &DDRB, &PINB, 1<<5 },
  [12] = { &PORTB, &DDRB, &PINB, 1<<6 },
  [13] = { &PORTB, &DDRB, &PINB, 1<<7 },
  [14] = { &PORTJ, &DDRJ, &PINJ, 1<<1 },
  [15] = { &PORTJ, &DDRJ, &PINJ, 1<<0 },
  [16] = { &PORTH, &DDRH, &PINH, 1<<1 },
  [17] = { &PORTH, &DDRH, &PINH, 1<<0 },
  [18] = { &PORTD, &DDRD, &PIND, 1<<3 },
  [19] = { &PORTD, &DDRD, &PIND, 1<<2 },
  [20] = { &PORTD, &DDRD, &PIND, 1<<1 },
  [21] = { &PORTD, &DDRD, &PIND, 1<<0 },
  [22] = { &PORTA, &DDRA, &PINA, 1<<0 },
  [23] = { &PORTA, &DDRA, &PINA, 1<<1 },
  [24] = { &PORTA, &DDRA, &PINA, 1<<2 },
  [25] = { &PORTA, &DDRA, &PINA, 1<<3 },
  [26] = { &PORTA, &DDRA, &PINA, 1<<4 },
  [27] = { &PORTA, &DDRA, &PINA, 1<<5 },
  [28] = { &PORTA, &DDRA, &PINA, 1<<6 },
  [29] = { &PORTA, &DDRA, &PINA, 1<<7 },
  [30] = { &PORTC, &DDRC, &PINC, 1<<7 },
  [31] = { &PORTC, &DDRC, &PINC, 1<<6 },
  [32] = { &PORTC, &DDRC, &PINC, 1<<5 },
  [33] = { &PORTC, &DDRC, &PINC, 1<<4 },
  [34] = { &PORTC, &DDRC, &PINC, 1<<3 },
  [35] = { &PORTC, &DDRC, &PINC, 1<<2 },
  [36] = { &PORTC, &DDRC, &PINC, 1<<1 },
  [37] = { &PORTC, &DDRC, &PINC, 1<<0 },
  [38] = { &PORTD, &DDRD, &PIND, 1<<7 },
  [39] = { &PORTG, &DDRG, &PING, 1<<2 },
  [40] = { &PORTG, &DDRG, &PING, 1<<1 },
  [41] = { &PORTG, &DDRG, &PING, 1<<0 },
  [42] = { &PORTL, &DDRL, &PINL, 1<<7 },
  [43] = { &PORTL, &DDRL, &PINL, 1<<6 },
  [44] = { &PORTL, &DDRL, &PINL, 1<<5 },
  [45] = { &PORTL, &DDRL, &PINL, 1<<4 },
  [46] = { &PORTL, &DDRL, &PINL, 1<<3 },
  [47] = { &PORTL, &DDRL, &PINL, 1<<2 },
  [48] = { &PORTL, &DDRL, &PINL, 1<<1 },
  [49] = { &PORTL, &DDRL, &PINL, 1<<0 },
  [50] = { &PORTB, &DDRB, &PINB, 1<<3 },
  [51] = { &PORTB, &DDRB, &PINB, 1<<2 },
  [52] = { &PORTB, &DDRB, &PINB, 1<<1 },
  [53] = { &PORTB, &DDRB, &PINB, 1<<0 }
#endif
};

static const uint8_t pwmsibs[ARIO_DPINS] = {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
  [3] = 11, [11] = 3,
  [5] = 6,  [6] = 5,
  [9] = 10, [10] = 9

#elif defined(ARDUIO_MEGA2560)
  [4] = 4,
  [11] = 12, [12] = 13, [13] = 11,
  [9] = 10, [10] = 9,
  [2] = 3, [3] = 5, [5] = 2,
  [6] = 7, [7] = 8, [8] = 6,
  [44] = 45, [45] = 46, [46] = 44
#endif
};

static uint16_t	ocr_scale(uint16_t v, uint16_t s);
static uint16_t	ocr_unscale(uint16_t v, uint16_t s);
static void	timer_prescale(uint8_t n, uint16_t s);
static uint8_t	timer_clksrc_extclk(uint16_t s);
static uint8_t	timer_clksrc_noextclk(uint16_t s);


/* --- ario_pin_init ------------------------------------------------------- */

void
ario_pin_init(void)
{
  uint8_t i;

  /* timers at max frequency, phase correct PWM */
  TCCR0A = (0 << WGM01) | (1 << WGM00);
  TCCR0B = (0 << WGM02);
  timer_prescale(0, 1);
  OCR0A = OCR0B = 0;

  TCCR1A = (1 << WGM11) | (0 << WGM10);
  TCCR1B = (1 << WGM13) | (0 << WGM12);
  timer_prescale(1, 1);
  ICR1 = 0xff;
  OCR1A = OCR1B = 0;

  TCCR2A = (0 << WGM21) | (1 << WGM20);
  TCCR2B = (0 << WGM22);
  timer_prescale(2, 1);
  OCR2A = OCR2B = 0;

#ifdef TCCR3A
  TCCR3A = (1 << WGM31) | (0 << WGM30);
  TCCR3B = (1 << WGM33) | (0 << WGM32);
  timer_prescale(3, 1);
  ICR3 = 0xff;
  OCR3A = OCR3B = 0;
#endif

#ifdef TCCR4A
  TCCR4A = (1 << WGM41) | (0 << WGM40);
  TCCR4B = (1 << WGM43) | (0 << WGM42);
  timer_prescale(4, 1);
  ICR4 = 0xff;
  OCR4A = OCR4B = 0;
#endif

#ifdef TCCR5A
  TCCR5A = (1 << WGM51) | (0 << WGM50);
  TCCR5B = (1 << WGM53) | (0 << WGM52);
  timer_prescale(5, 1);
  ICR5 = 0xff;
  OCR5A = OCR5B = 0;
#endif

  /* ADC, AVCC ref, single conversion, 125kHz (~100µs/chan) */
  ADMUX = (0 << REFS1) | (1 << REFS0);
  DIDR0 =
    (1 << ADC5D) | (1 << ADC4D) | (1 << ADC3D) |
    (1 << ADC2D) | (1 << ADC1D) | (1 << ADC0D);
#ifdef ARDUIO_MEGA2560
  DIDR0 |= (1 << ADC7D) | (1 << ADC6D);
  DIDR2 =
    (1 << ADC15D) | (1 << ADC14D) | (1 << ADC13D) | (1 << ADC12D) |
    (1 << ADC11D) | (1 << ADC10D) | (1 << ADC9D) | (1 << ADC8D);
#endif
  ADCSRB = (0 << ADTS2) | (0 << ADTS1) | (0 << ADTS0);
  ADCSRA =
    (1 << ADEN) | (1 << ADSC) | (0 << ADATE) | (1 << ADIF) | (1 << ADIE) |
    (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);


  /* input pins by default */
  for(i = 0; i < ARIO_DPINS; i++)
    ario_dpin_dir(i, ARIO_IN);
}


/* --- ario_dpin_dir ------------------------------------------------------- */

void
ario_dpin_dir(uint8_t n, enum ario_dir dir)
{
  if (n >= ARIO_DPINS || n < 2) return;

  /* configure PWM first */
  switch(dir) {
    case ARIO_PWM_31kHz:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case 5:  case 6: timer_prescale(0, 1); break;
        case 9: case 10: timer_prescale(1, 1); ICR1 = 0xff; break;
        case 3: case 11: timer_prescale(2, 1); break;
#elif defined(ARDUIO_MEGA2560)
        case 4:                    timer_prescale(0, 1); break;
        case 11: case 12: case 13: timer_prescale(1, 1); ICR1 = 0xff; break;
        case 9: case 10:           timer_prescale(2, 1); break;
        case 2: case 3: case 5:    timer_prescale(3, 1); ICR3 = 0xff; break;
        case 6: case 7: case 8:    timer_prescale(4, 1); ICR4 = 0xff; break;
        case 44: case 45: case 46: timer_prescale(5, 1); ICR5 = 0xff; break;
#endif
      }
      break;

    case ARIO_PWM_122Hz:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case 5:  case 6: timer_prescale(0, 256); break;
        case 9: case 10: timer_prescale(1, 1); ICR1 = 0xffff; break;
        case 3: case 11: timer_prescale(2, 256); break;
#elif defined(ARDUIO_MEGA2560)
        case 4:                    timer_prescale(0, 256); break;
        case 11: case 12: case 13: timer_prescale(1, 1); ICR1 = 0xffff; break;
        case 9: case 10:           timer_prescale(2, 256); break;
        case 2: case 3: case 5:    timer_prescale(3, 1); ICR3 = 0xffff; break;
        case 6: case 7: case 8:    timer_prescale(4, 1); ICR4 = 0xffff; break;
        case 44: case 45: case 46: timer_prescale(5, 1); ICR5 = 0xffff; break;
#endif
      }
      break;

    case ARIO_PWM_31Hz:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case 5:  case 6: timer_prescale(0, 1024); break;
        case 9: case 10: timer_prescale(1, 8); ICR1 = 0x7fff; break;
        case 3: case 11: timer_prescale(2, 1024); break;
#elif defined(ARDUIO_MEGA2560)
        case 4:                    timer_prescale(0, 1024); break;
        case 11: case 12: case 13: timer_prescale(1, 8); ICR1 = 0x7fff; break;
        case 9: case 10:           timer_prescale(2, 1024); break;
        case 2: case 3: case 5:    timer_prescale(3, 8); ICR3 = 0x7fff; break;
        case 6: case 7: case 8:    timer_prescale(4, 8); ICR4 = 0x7fff; break;
        case 44: case 45: case 46: timer_prescale(5, 8); ICR5 = 0x7fff; break;
#endif
      }
      break;

    default: break;
  }

  switch(dir) {
    case ARIO_IN:
    case ARIO_INPU:
    case ARIO_OUT:
    case ARIO_WS28XX:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case  3: TCCR2A &= ~((1 << COM2B1) | (1 << COM2B0)); break;
        case  5: TCCR0A &= ~((1 << COM0B1) | (1 << COM0B0)); break;
        case  6: TCCR0A &= ~((1 << COM0A1) | (1 << COM0A0)); break;
        case  9: TCCR1A &= ~((1 << COM1A1) | (1 << COM1A0)); break;
        case 10: TCCR1A &= ~((1 << COM1B1) | (1 << COM1B0)); break;
        case 11: TCCR2A &= ~((1 << COM2A1) | (1 << COM2A0)); break;
#elif defined(ARDUIO_MEGA2560)
        case  2: TCCR3A &= ~((1 << COM3B1) | (1 << COM3B0)); break;
        case  3: TCCR3A &= ~((1 << COM3C1) | (1 << COM3C0)); break;
        case  4: TCCR0A &= ~((1 << COM0B1) | (1 << COM0B0)); break;
        case  5: TCCR3A &= ~((1 << COM3A1) | (1 << COM3A0)); break;
        case  6: TCCR4A &= ~((1 << COM4A1) | (1 << COM4A0)); break;
        case  7: TCCR4A &= ~((1 << COM4B1) | (1 << COM4B0)); break;
        case  8: TCCR4A &= ~((1 << COM4C1) | (1 << COM4C0)); break;
        case  9: TCCR2A &= ~((1 << COM2B1) | (1 << COM2B0)); break;
        case 10: TCCR2A &= ~((1 << COM2A1) | (1 << COM2A0)); break;
        case 11: TCCR1A &= ~((1 << COM1A1) | (1 << COM1A0)); break;
        case 12: TCCR1A &= ~((1 << COM1B1) | (1 << COM1B0)); break;
        case 13: TCCR1A &= ~((1 << COM1C1) | (1 << COM1C0)); break;
        case 44: TCCR5A &= ~((1 << COM5C1) | (1 << COM5C0)); break;
        case 45: TCCR5A &= ~((1 << COM5B1) | (1 << COM5B0)); break;
        case 46: TCCR5A &= ~((1 << COM5A1) | (1 << COM5A0)); break;
#endif
      }
      break;

    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case  3: TCCR2A |= (1 << COM2B1) | (0 << COM2B0); break;
        case  5: TCCR0A |= (1 << COM0B1) | (0 << COM0B0); break;
        case  6: TCCR0A |= (1 << COM0A1) | (0 << COM0A0); break;
        case  9: TCCR1A |= (1 << COM1A1) | (0 << COM1A0); break;
        case 10: TCCR1A |= (1 << COM1B1) | (0 << COM1B0); break;
        case 11: TCCR2A |= (1 << COM2A1) | (0 << COM2A0); break;
#elif defined(ARDUIO_MEGA2560)
        case  2: TCCR3A |= (1 << COM3B1) | (0 << COM3B0); break;
        case  3: TCCR3A |= (1 << COM3C1) | (0 << COM3C0); break;
        case  4: TCCR0A |= (1 << COM0B1) | (0 << COM0B0); break;
        case  5: TCCR3A |= (1 << COM3A1) | (0 << COM3A0); break;
        case  6: TCCR4A |= (1 << COM4A1) | (0 << COM4A0); break;
        case  7: TCCR4A |= (1 << COM4B1) | (0 << COM4B0); break;
        case  8: TCCR4A |= (1 << COM4C1) | (0 << COM4C0); break;
        case  9: TCCR2A |= (1 << COM2B1) | (0 << COM2B0); break;
        case 10: TCCR2A |= (1 << COM2A1) | (0 << COM2A0); break;
        case 11: TCCR1A |= (1 << COM1A1) | (0 << COM1A0); break;
        case 12: TCCR1A |= (1 << COM1B1) | (0 << COM1B0); break;
        case 13: TCCR1A |= (1 << COM1C1) | (0 << COM1C0); break;
        case 44: TCCR5A |= (1 << COM5C1) | (0 << COM5C0); break;
        case 45: TCCR5A |= (1 << COM5B1) | (0 << COM5B0); break;
        case 46: TCCR5A |= (1 << COM5A1) | (0 << COM5A0); break;
#endif

        default: return;
      }
      break;
  }

  /* configure direction */
  switch(dir) {
    case ARIO_IN:
      *dpinsreg[n].port &= ~dpinsreg[n].bit;
      *dpinsreg[n].ddr &= ~dpinsreg[n].bit;
      break;

    case ARIO_INPU:
      *dpinsreg[n].ddr &= ~dpinsreg[n].bit;
      *dpinsreg[n].port |= dpinsreg[n].bit;
      break;

    case ARIO_OUT:
    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
    case ARIO_WS28XX:
      *dpinsreg[n].ddr |= dpinsreg[n].bit;
      break;
  }

  /* update the frequency of coupled PWM pins */
  switch(dir) {
    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      if (pwmsibs[n]) {
        uint8_t s;
        for(s = pwmsibs[n]; s != n; s = pwmsibs[s]) {
          if (ario_dpins[s] > ARIO_OUT &&
              ario_dpins[s] < ARIO_WS28XX)
            ario_dpins[s] = dir;
        }
      }
      break;

    default: break;
  }

  ario_dpins[n] = dir;
}


/* --- ario_dpin_out ------------------------------------------------------- */

void
ario_dpin_out(uint8_t n, uint16_t v)
{
  if (n >= ARIO_DPINS || n < 2) return;

  switch(ario_dpins[n]) {
    case ARIO_IN:
    case ARIO_INPU:
      break;

    case ARIO_OUT:
      if (v)
        *dpinsreg[n].port |= dpinsreg[n].bit;
      else
        *dpinsreg[n].port &= ~dpinsreg[n].bit;
      break;

    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case  3: OCR2B = v >> 8; break;
        case  5: OCR0B = v >> 8; break;
        case  6: OCR0A = v >> 8; break;
        case  9: OCR1A = ocr_scale(v, ICR1); break;
        case 10: OCR1B = ocr_scale(v, ICR1); break;
        case 11: OCR2A = v >> 8; break;

#elif defined(ARDUIO_MEGA2560)
        case  2: OCR3B = ocr_scale(v, ICR3); break;
        case  3: OCR3C = ocr_scale(v, ICR3); break;
        case  4: OCR0B = v >> 8; break;
        case  5: OCR3A = ocr_scale(v, ICR3); break;
        case  6: OCR4A = ocr_scale(v, ICR4); break;
        case  7: OCR4B = ocr_scale(v, ICR4); break;
        case  8: OCR4C = ocr_scale(v, ICR4); break;
        case  9: OCR2B = v >> 8; break;
        case 10: OCR2A = v >> 8; break;
        case 11: OCR1A = ocr_scale(v, ICR1); break;
        case 12: OCR1B = ocr_scale(v, ICR1); break;
        case 13: OCR1C = ocr_scale(v, ICR1); break;
        case 44: OCR5C = ocr_scale(v, ICR5); break;
        case 45: OCR5B = ocr_scale(v, ICR5); break;
        case 46: OCR5A = ocr_scale(v, ICR5); break;
#endif

        default: return;
      }
      break;

    case ARIO_WS28XX: {
      uint8_t i;
      uint8_t led = v >> 12;

      ario_wspins[n][led] = v & 0xfff;
      ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
        for(i = 0; i <= led; i++)
          ws28xx_set(dpinsreg[n].port, dpinsreg[n].bit, ario_wspins[n][i]);
      }
      _delay_us(50);
      break;
    }
  }
}


/* --- ario_dpin_in -------------------------------------------------------- */

uint16_t
ario_dpin_in(uint8_t n)
{
  if (n >= ARIO_DPINS || n < 2) return 0;

  switch(ario_dpins[n]) {
    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      switch(n) {
#if defined(ARDUIO_UNO) || defined(ARDUIO_NANO)
        case  3: return OCR2B << 8;
        case  5: return OCR0B << 8;
        case  6: return OCR0A << 8;
        case  9: return ocr_unscale(OCR1A, ICR1);
        case 10: return ocr_unscale(OCR1B, ICR1);
        case 11: return OCR2A << 8;

#elif defined(ARDUIO_MEGA2560)
        case  2: return ocr_unscale(OCR3B, ICR3);
        case  3: return ocr_unscale(OCR3C, ICR3);
        case  4: return OCR0B << 8;
        case  5: return ocr_unscale(OCR3A, ICR3);
        case  6: return ocr_unscale(OCR4A, ICR4);
        case  7: return ocr_unscale(OCR4B, ICR4);
        case  8: return ocr_unscale(OCR4C, ICR4);
        case  9: return OCR2B << 8;
        case 10: return OCR2A << 8;
        case 11: return ocr_unscale(OCR1A, ICR1);
        case 12: return ocr_unscale(OCR1B, ICR1);
        case 13: return ocr_unscale(OCR1C, ICR1);
        case 44: return ocr_unscale(OCR5C, ICR5);
        case 45: return ocr_unscale(OCR5B, ICR5);
        case 46: return ocr_unscale(OCR5A, ICR5);
#endif
      }
      /*FALLTHROUGH*/
    case ARIO_IN:
    case ARIO_INPU:
    case ARIO_OUT:
      return ((*dpinsreg[n].pin) & dpinsreg[n].bit) ? 0xffff : 0;

    case ARIO_WS28XX:
      return ario_wspins[n][0];
  }

  return 0;
}


/* --- ario_apin_in -------------------------------------------------------- */

uint16_t
ario_apin_in(uint8_t n)
{
  if (n >= ARIO_APINS) return 0;
  return ario_apins[n];
}


/* --- ario_ddir_state ----------------------------------------------------- */

void
ario_ddir_state(void)
{
  uint8_t dir[1 + ARIO_DPINS];
  uint8_t i;

  dir[0] = 'D';
  for(i = 0; i < sizeof(ario_dpins)/sizeof(ario_dpins[0]); i++)
    dir[i+1] = ario_dpins[i];

  ario_log_buffer(dir, sizeof(dir)/sizeof(dir[0]));
}


/* --- ario_dpin_state ----------------------------------------------------- */

void
ario_dpin_state(void)
{
  uint8_t in[1 + 2 * ARIO_DPINS];
  uint16_t v;
  uint8_t i;

  in[0] = 'I';
  for(i = 0; i < ARIO_DPINS; i++) {
    v = ario_dpin_in(i);
    in[2*i+1] = v >> 8;
    in[2*i+2] = v & 0xff;
  }

  ario_log_buffer(in, sizeof(in));
}


/* --- ario_apin_state ----------------------------------------------------- */

void
ario_apin_state(void)
{
  uint8_t in[1 + 2 * ARIO_APINS];
  uint16_t v;
  uint8_t i;

  in[0] = 'A';
  for(i = 0; i < ARIO_APINS; i++) {
    v = ario_apins[i];
    in[2*i+1] = v >> 8;
    in[2*i+2] = v & 0xff;
  }

  ario_log_buffer(in, sizeof(in));
}


/* --- ocr_scale ----------------------------------------------------------- */

static uint16_t
ocr_scale(uint16_t v, uint16_t s)
{
  switch(s) {
    case 0xffff: return v;
    case 0x7fff: return v >> 1;
    case   0xff: return v >> 8;
  }
  return 0;
}

static uint16_t
ocr_unscale(uint16_t v, uint16_t s)
{
  switch(s) {
    case 0xffff: return v;
    case 0x7fff: return v << 1;
    case   0xff: return v << 8;
  }
  return 0;
}

/* --- timer_prescale ------------------------------------------------------ */

static void
timer_prescale(uint8_t n, uint16_t s)
{
  switch(n) {
    case 0: TCCR0B = (TCCR0B & ~7) | timer_clksrc_extclk(s); break;
    case 1: TCCR1B = (TCCR1B & ~7) | timer_clksrc_extclk(s); break;
    case 2: TCCR2B = (TCCR2B & ~7) | timer_clksrc_noextclk(s); break;

#ifdef TCCR3B
    case 3: TCCR3B = (TCCR3B & ~7) | timer_clksrc_extclk(s); break;
#endif
#ifdef TCCR4B
    case 4: TCCR4B = (TCCR4B & ~7) | timer_clksrc_extclk(s); break;
#endif
#ifdef TCCR5B
    case 5: TCCR5B = (TCCR5B & ~7) | timer_clksrc_extclk(s); break;
#endif
  }
}

static uint8_t
timer_clksrc_extclk(uint16_t s)
{
  switch(s) {
    case    0: return 0;
    case    1: return 1;
    case    8: return 2;
    case   64: return 3;
    case  256: return 4;
    case 1024: return 5;

    default: return 0;
  }
}

static uint8_t
timer_clksrc_noextclk(uint16_t s)
{
  switch(s) {
    case    0: return 0;
    case    1: return 1;
    case    8: return 2;
    case   32: return 3;
    case   64: return 4;
    case  128: return 5;
    case  256: return 6;
    case 1024: return 7;

    default: return 0;
  }
}


/* --- ADC_vect ------------------------------------------------------------ */

ISR(ADC_vect)
{
  uint8_t adc_channel = ADMUX & 0x7;
#ifdef ARDUIO_MEGA2560
  adc_channel |= ADCSRB & 0x8;
#endif

  /* read ADC */
  ario_apins[adc_channel++] = ADC << 6;
  if (adc_channel >= ARIO_APINS) adc_channel = 0;
  ADMUX = (ADMUX & 0xe0) | (adc_channel & 0x7);
#ifdef ARDUIO_MEGA2560
  ADCSRB = (ADCSRB & ~0x8) | (adc_channel & 0x8);
#endif

  ADCSRA |= (1 << ADSC);
}
